/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package apiclient

import (
	"fmt"
	"os"
	"path/filepath"
	"time"

	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/varnish"
	"code.uplex.de/uplex-varnish/k8s-vcl-reloader/pkg/varnish/vcl"

	"github.com/sirupsen/logrus"

	api_v1 "k8s.io/api/core/v1"
	core_v1_listers "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
)

const vclSpecErr = "VCLspecError"

// SpecUpdaterConfig represents the user configuration for k8s
// resources that are considered for a VCL update.
type SpecUpdaterConfig struct {
	SvcKeys         []string
	Generated       []string
	Main            string
	CfgMapName      string
	CfgMapNamespace string
	VersionKey      string
	VCLRoot         string
}

// SpecUpdater encapsulates updating the VCL spec from the current
// state of the k8s cluster.
type SpecUpdater struct {
	log     *logrus.Logger
	cfg     SpecUpdaterConfig
	vcls    map[string]time.Time
	genvcls map[string]time.Time
	vc      *varnish.Controller
	svc     core_v1_listers.ServiceLister
	endp    core_v1_listers.EndpointsLister
	cfgMap  core_v1_listers.ConfigMapLister
}

// NewSpecUpdater creates a SpecUpdater.
func NewSpecUpdater(
	log *logrus.Logger,
	cfg SpecUpdaterConfig,
	vc *varnish.Controller,
	listers *Listers,
) *SpecUpdater {
	genvcls := make(map[string]time.Time, len(cfg.Generated))
	for _, gen := range cfg.Generated {
		genvcls[gen] = time.Time{}
	}
	return &SpecUpdater{
		log:     log,
		cfg:     cfg,
		genvcls: genvcls,
		vc:      vc,
		endp:    listers.endp,
		svc:     listers.svc,
		cfgMap:  listers.cfgMap,
	}
}

func getEndpSpec(addr api_v1.EndpointAddress, port int32) vcl.Endpoint {
	ep := vcl.Endpoint{
		IP:   addr.IP,
		Port: port,
	}
	ep.Namespace = addr.TargetRef.Namespace
	ep.UID = string(addr.TargetRef.UID)
	ep.ResourceVersion = addr.TargetRef.ResourceVersion
	return ep
}

func (updater *SpecUpdater) getSvcSpec(
	svc *api_v1.Service) (vclSvc vcl.Service, err error) {

	updater.log.Debugf("Updating Service %s/%s spec from cluster config",
		svc.Namespace, svc.Name)

	vclSvc = vcl.Service{}
	eps, err := updater.endp.Endpoints(svc.Namespace).Get(svc.Name)
	if err != nil {
		return
	}

	vclSvc.Namespace = svc.Namespace
	vclSvc.UID = string(svc.UID)
	vclSvc.ResourceVersion = svc.ResourceVersion
	vclSvc.Endpoints = make(map[string]vcl.Endpoint)
	for _, subset := range eps.Subsets {
		if len(subset.Ports) != 1 {
			// XXX ???
			err = fmt.Errorf("Endpoint %s/%s has %d ports",
				eps.Namespace, eps.Name, len(subset.Ports))
			return
		}
		port := subset.Ports[0].Port
		for _, addr := range subset.Addresses {
			vclSvc.Endpoints[addr.TargetRef.Name] =
				getEndpSpec(addr, port)
		}
	}
	return
}

func (updater *SpecUpdater) getSpec() (vcl.Spec, error) {
	updater.log.Info("Updating VCL spec from cluster config")

	spec := vcl.Spec{
		Main:    updater.cfg.Main,
		Created: time.Now(),
	}

	svcsSpec := make(map[string]vcl.Service)
	for _, svcKey := range updater.cfg.SvcKeys {
		ns, name, err := cache.SplitMetaNamespaceKey(svcKey)
		if err != nil {
			return spec, err
		}
		svc, err := updater.svc.Services(ns).Get(name)
		if err != nil {
			return spec, err
		}
		// XXX same name in different namespaces
		svcsSpec[svc.Name], err = updater.getSvcSpec(svc)
		if err != nil {
			return spec, err
		}
	}
	if len(svcsSpec) != len(updater.cfg.SvcKeys) {
		var svcsNotFound []string
		for _, key := range updater.cfg.SvcKeys {
			_, name, err := cache.SplitMetaNamespaceKey(key)
			if err != nil {
				return spec, err
			}
			if _, exists := svcsSpec[name]; !exists {
				svcsNotFound = append(svcsNotFound, name)
			}
		}
		return spec, fmt.Errorf("Services not found, no update: %s",
			svcsNotFound)
	}
	spec.Services = svcsSpec

	cfgMap, err := updater.cfgMap.ConfigMaps(updater.cfg.CfgMapNamespace).
		Get(updater.cfg.CfgMapName)
	if err != nil {
		return spec, err
	}
	version := ""
	if updater.cfg.VersionKey != "" {
		if v, exists := cfgMap.Data[updater.cfg.VersionKey]; exists {
			version = v
		}
	}

	spec.Version = version
	spec.CfgMap.Name = cfgMap.Name
	spec.CfgMap.Namespace = cfgMap.Namespace
	spec.CfgMap.Version = version
	spec.CfgMap.UID = string(cfgMap.UID)
	spec.CfgMap.ResourceVersion = cfgMap.ResourceVersion

	return spec, nil
}

func getKeySet(cfgMap *api_v1.ConfigMap) map[string]struct{} {
	keys := make(map[string]struct{})
	if cfgMap.Data != nil {
		for k := range cfgMap.Data {
			keys[k] = struct{}{}
		}
	}
	if cfgMap.BinaryData != nil {
		for k := range cfgMap.BinaryData {
			keys[k] = struct{}{}
		}
	}
	return keys
}

// VCLChanged returns true if the VCL files denoted by the keys in the
// ConfigMap have changed since its last update.
//
// Return true if the ConfigMap has never been synced. Return false if
// the ConfigMap has no keys (then there are no VCL files to reload).
//
// If a key has been added to the ConfigMap since the last update,
// return true if the corresponding file exists, false if the file is
// not found. If a key has been deleted, return true if the
// corresponding file has also been deleted, false if the file still
// exists.
//
// Otherwise, return true if and only if the file modification times
// have changed since the last update.
//
// Return any error related to accessing files, except for file not
// found errors that are expected after key addition or deletion.
func (updater *SpecUpdater) VCLChanged(cfgMap *api_v1.ConfigMap) (bool, error) {
	if cfgMap.Data == nil && cfgMap.BinaryData == nil {
		updater.log.Warnf("ConfigMap %s/%s has no data",
			cfgMap.Namespace, cfgMap.Name)
		return false, nil
	}

	files := getKeySet(cfgMap)
	if len(files) == 0 {
		updater.log.Warnf("ConfigMap %s/%s has no data",
			cfgMap.Namespace, cfgMap.Name)
		return false, nil
	}

	if updater.vcls == nil || len(updater.vcls) == 0 {
		updater.log.Trace("Never synced")
		return true, nil
	}

	for k := range files {
		if _, ok := updater.vcls[k]; !ok {
			updater.log.Trace("New key: ", k)
			path := filepath.Join(updater.cfg.VCLRoot, k)
			_, err := os.Stat(path)
			if err == nil {
				updater.log.Trace("File added: ", k)
				return true, nil
			}
			if os.IsNotExist(err) {
				updater.log.Trace("File not added: ", k)
				return false, nil
			}
			return false, err
		}
	}
	for k := range updater.vcls {
		if _, ok := files[k]; !ok {
			updater.log.Trace("Key deleted: ", k)
			path := filepath.Join(updater.cfg.VCLRoot, k)
			_, err := os.Stat(path)
			if err == nil {
				updater.log.Trace("File not deleted: ", k)
				return false, nil
			}
			if os.IsNotExist(err) {
				updater.log.Trace("File deleted: ", k)
				return true, nil
			}
			return false, err
		}
	}

	for file := range files {
		path := filepath.Join(updater.cfg.VCLRoot, file)
		info, err := os.Stat(path)
		if err != nil {
			return false, err
		}
		if updater.vcls[file] != info.ModTime() {
			updater.log.Trace(file, " new mtime ", info.ModTime())
			return true, nil
		}
	}
	return false, nil
}

func (updater *SpecUpdater) genVCLChanged() (bool, error) {
	for path, mtime := range updater.genvcls {
		info, err := os.Stat(path)
		if err != nil {
			if os.IsNotExist(err) {
				updater.log.Warn("Generated VCL not found: ",
					path)
				continue
			}
			return false, err
		}
		if mtime != info.ModTime() {
			updater.log.Trace(path, " new mtime ", info.ModTime())
			return true, nil
		}
	}
	return false, nil
}

// ClearVCL resets the record of VCL files and their modification
// times. The record of VCL files generated from template is
// unaffected.
func (updater *SpecUpdater) ClearVCL() {
	updater.vcls = nil
}

func (updater *SpecUpdater) updatedVCLMap(
	cfgMap *api_v1.ConfigMap) (vclMap map[string]time.Time, err error) {

	if cfgMap.Data == nil && cfgMap.BinaryData == nil {
		updater.log.Warnf("ConfigMap %s/%s has no data",
			cfgMap.Namespace, cfgMap.Name)
		return
	}

	files := getKeySet(cfgMap)
	vclMap = make(map[string]time.Time, len(files))
	for file := range files {
		path := filepath.Join(updater.cfg.VCLRoot, file)
		info, err := os.Stat(path)
		if err != nil {
			return nil, err
		}
		vclMap[file] = info.ModTime()
	}
	return
}

func (updater *SpecUpdater) updatedGenVCLMap() (
	map[string]time.Time,
	error,
) {
	vclMap := make(map[string]time.Time, len(updater.genvcls))
	for path := range updater.genvcls {
		info, err := os.Stat(path)
		if err != nil {
			if os.IsNotExist(err) {
				updater.log.Warn("Generated VCL not found: ",
					path)
				continue
			}
			return nil, err
		}
		vclMap[path] = info.ModTime()
	}
	return vclMap, nil
}

func (updater *SpecUpdater) latestVCLMtime() time.Time {
	latest := time.Time{}
	for _, t := range updater.vcls {
		if t.After(latest) {
			latest = t
		}
	}
	for _, t := range updater.genvcls {
		if t.After(latest) {
			latest = t
		}
	}
	return latest
}

// Update generates a new configuration from the current state of the
// k8s cluster, and loads it as a new instance of VCL.
func (updater *SpecUpdater) Update() error {
	updater.log.Debug("Current VCL file modification times: ",
		updater.vcls, updater.genvcls)
	updater.log.Info("Current latest known VCL file modification time: ",
		updater.latestVCLMtime())
	spec, err := updater.getSpec()
	if err != nil {
		updater.vc.ErrorEvt(vclSpecErr, "Cannot create VCL spec: %+v",
			err)
		return err
	}
	cfgMap, err := updater.cfgMap.ConfigMaps(spec.CfgMap.Namespace).
		Get(spec.CfgMap.Name)
	if err != nil {
		updater.log.Errorf("Cannot retrieve ConfigMap %s/%s: %v",
			spec.CfgMap.Namespace, spec.CfgMap.Name, err)
		return err
	}
	genVCLChanged, err := updater.genVCLChanged()
	if err != nil {
		updater.log.Errorf("Error checking generated VCL mtimes: %v",
			err)
		return err
	}
	if !updater.vc.HasConfigMap(spec.CfgMap) {
		if changed, err := updater.VCLChanged(cfgMap); err != nil {
			return err
		} else if !genVCLChanged && !changed {
			updater.log.Warnf("ConfigMap %s/%s has changed but "+
				"VCL files have not been updated, not "+
				"updating Varnish config",
				spec.CfgMap.Namespace, spec.CfgMap.Name)
			return nil
		}
	}
	if updater.vc.HasConfig(spec) {
		updater.log.Infof("Varnish VCL config is up to date")
		return nil
	}
	updatedMap, err := updater.updatedVCLMap(cfgMap)
	if err != nil {
		updater.log.Errorf("Cannot update VCL files and modification "+
			"times: %v", err)
		return err
	}
	updatedGenMap, err := updater.updatedGenVCLMap()
	if err != nil {
		updater.log.Errorf("Cannot update generated VCL mtimes: %v",
			err)
		return err
	}
	if err = updater.vc.Update(&spec); err != nil {
		updater.vc.ErrorEvt(varnish.UpdateErr, "VCL update failed: %s",
			err)
		return err
	}
	// XXX generate Info event, inc counter(?)
	updater.vcls = updatedMap
	updater.genvcls = updatedGenMap
	updater.log.Debug("Updated VCL file modification times: ", updater.vcls,
		updater.genvcls)
	updater.log.Info("Updated latest known VCL file modification time: ",
		updater.latestVCLMtime())
	updater.log.Info("Varnish update succeeded")
	return nil
}
